/*	BitmapStream.cpp
 */

#include "BitmapStream.h"
#include <Bitmap.h>
#define DEBUG 1
#include <Debug.h>



BitmapStream::BitmapStream(
	BBitmap *		bitmap)
{
	fMap = bitmap;
	fDetached = false;
	fPosition = 0;
	fSize = 0;
	/*	Extract header if needed */
	if (fMap) {
		fHeader.magic = DATA_BITMAP;
		fHeader.bounds = fMap->Bounds();
		fHeader.rowBytes = fMap->BytesPerRow();
		fHeader.colors = fMap->ColorSpace();
		fHeader.dataSize = (fHeader.bounds.Height()+1)*fHeader.rowBytes;
		fSize = sizeof(DATABitmap)+fHeader.dataSize;
	}
}

BitmapStream::~BitmapStream()
{
	if (fMap && !fDetached)
		delete fMap;
}

long
BitmapStream::Read(
	void *			buffer,
	long &			size)
{
	if (!fMap)
		return B_ERROR;
	if (!size)
		return B_NO_ERROR;
	if (size < 0)
		return B_ERROR;
	if (fPosition >= fSize)
		return B_ERROR;

	long toRead;
	void *source;

	if (fPosition < sizeof(DATABitmap)) {
		toRead = sizeof(DATABitmap)-fPosition;
		source = ((char *)&fHeader)+fPosition;
	} else {
		toRead = fSize-fPosition;
		source = ((char *)fMap->Bits())+fPosition-sizeof(DATABitmap);
	}
	if (toRead > size)
		toRead = size;
	memcpy(buffer, source, toRead);
	fPosition += toRead;
	size = toRead;
	return B_NO_ERROR;
}

long
BitmapStream::Write(
	const void *	data,
	long			size)
{
	if (!size)
		return B_NO_ERROR;
	if (size < 0)
		return B_ERROR;
	while (size > 0) {
		long toWrite;
		void *dest;
		/*	We depend on writing the header separately in detecting changes to it */
		if (fPosition < sizeof(DATABitmap)) {
			toWrite = sizeof(DATABitmap)-fPosition;
			dest = ((char *)&fHeader)+fPosition;
		} else {
			toWrite = fHeader.dataSize-fPosition+sizeof(DATABitmap);
			dest = ((char *)fMap->Bits())+fPosition-sizeof(DATABitmap);
		}
		if (toWrite > size)
			toWrite = size;
		if (!toWrite && size)	//	i e we've been told to write too much
			return B_BAD_VALUE;
		memcpy(dest, data, toWrite);
		fPosition += toWrite;
		data = (char *)data+toWrite;
		size -= toWrite;
		if (fPosition > fSize)
			fSize = fPosition;
		/*	If we change the header, the rest goes */
		if (fPosition == sizeof(DATABitmap)) {
			if (fMap && ((fMap->Bounds() != fHeader.bounds) ||
					(fMap->ColorSpace() != fHeader.colors) ||
					(fMap->BytesPerRow() != fHeader.rowBytes))) {
				if (!fDetached)	//	if someone detached, we don't delete
					delete fMap;
				fMap = NULL;
			}
			if (!fMap) {
				if ((fHeader.bounds.left > 0.0) || (fHeader.bounds.top > 0.0))
					DEBUGGER("non-origin bounds!");
				fMap = new BBitmap(fHeader.bounds, fHeader.colors);
				if (fMap->BytesPerRow() != fHeader.rowBytes) {
					return B_MISMATCHED_VALUES;
				}
			}
		}
	}
	return B_NO_ERROR;
}

long
BitmapStream::Seek(	//	returns 0 for success
	long			position)
{
	if (position < 0)
		return B_BAD_VALUE;
	if (position > fSize)
		return B_BAD_VALUE;
	fPosition = position;
	return B_NO_ERROR;
}

long
BitmapStream::Position()
{
	return fPosition;
}

long
BitmapStream::Size()
{
	return fSize;
}

long
BitmapStream::SetSize(	//	returns 0 for success
	long			size)
{
	if (size < 0)
		return B_BAD_VALUE;
	if (fMap && (size > fHeader.dataSize+sizeof(DATABitmap)))
		return B_BAD_VALUE;
	/*	Problem:
	 *	What if someone calls SetSize() before writing the header, so we don't know what 
	 *	bitmap to create?
	 *	Solution:
	 *	We assume people will write the header before any data. Else they shouldn'e use 
	 *	this class. So SetSize() is really not going to do anything.
	 */
	if (fMap)	//	if we checked that the size was OK
		fSize = size;
	return B_NO_ERROR;
}

long
BitmapStream::DetachBitmap(
	BBitmap * &	outBitmap)
{
	outBitmap = NULL;
	if (!fMap)
		return B_ERROR;
	if (fDetached)
		return B_ERROR;
	fDetached = true;
	outBitmap = fMap;
	return B_NO_ERROR;
}

