/*	DataVideo.h
 *	Copyright  1996 Jon Watte
 *
 *	This format is the "standard" interchange format for streaming media, such as video, 
 *	animation and/or sound interleaved, in the context of Datatypes.lib. It is designed to 
 *	be easy to read and write without seeking, more than for performance or capabilities 
 *	- for instance, no compression.
 *	We also do NOT use DATA_BITMAP or DATA_SOUND chunks, because those do 
 *	not conform to the magic-headersize-datasize-position header layout, and also use 
 *	floating point numbers in headers which may or may not be portable across platforms.
 *	The well-defined byte order is big endian (68K, PPC) at least for now.
 *
 *	To use the classes defined herein, you will have to include the source file DataVideo.cpp 
 *	in your application; it's not part of Datatypes.lib.
 */

#pragma once


#include "DStream.h"


enum { MEDIA_START_MAGIC = 'mhi!' };
struct MediaStreamHeader {
	ulong		magic;			/*	MEDIA_START_MAGIC	*/
	long		headersize;		/*	size of this header (20 for now)	*/
	long		datasize;		/*	size of data (not including header)	*/
	long		ticksstart; 	/*	media clock when file was started
								 *	Typically you'd want it as 0, but other values are OK
								 *	Even though each chunk has its own timestamp, they are 
								 *	expected to never be decreasing	*/
	long		tickspersec;	/*	speed of stream
								 *	The convention is to use one of the values 1000, 44100 or 48000
								 *	and ALWAYS use the sound sample rate if there is sound...	*/
/*	ulong		tagsrequired[datasize/sizeof(ulong)];
 *	here, you write each tag that you think you will use in the file.
 */
};

/*	MEDIA_BITMAP_HEADER is a special bitmap header that contains information about 
 *	the coming bitmap frames. It defines the preferred color space, the total size of the frame, 
 *	and the rowbytes to use at that size.
 *	This is used because the first BITMAP_CHUNK may not encompass all of the desired 
 *	frame size.
 *	MEDIA_BITMAP_HEADER uses the MediaBitmapHeader struct for this purpose, but 
 *	contains no data after the header.
 */
enum { MEDIA_BITMAP_HEADER = 'bmbh' };
/*	MEDIA_BITMAP_CHUNK describes a sub-part of the total image to be displayed.
*	It can be a sub-part of the desired image or frame.
*/
enum { MEDIA_BITMAP_CHUNK = 'mbmc' };
enum {	/*	encodings for the "space" header value	*/
	MEDIA_8_BIT_FIXED = 8,
	MEDIA_8_BIT_GRAY  = 9,
	MEDIA_8_BIT_PALETTE = 10,
	MEDIA_32_BIT_BGRA = 32,
	MEDIA_32_BIT_ARGB = 32+256,
	MEDIA_16_BIT_BGRA = 16,
	MEDIA_16_BIT_ARGB = 16+256
};
struct MediaBitmapHeader {
	ulong		magic;		/*	MEDIA_BITMAP_CHUNK	*/
	ulong		hdrsize;	/*	(24 for now)	*/
	ulong		datasize;	/*	same as ysize*rowbytes	*/
	long		position;	/*	media clock when to play it	*/
	ushort		space;		/*	this would NOT be color_space values
							 *	see the enum above for values
							 */
	ushort		xpos;		/*	Where within bitmap to place data	*/
	ushort		ypos;		/*	Where within bitmap to place data	*/
	ushort		xsize;		/*	number of pixels	*/
	ushort		ysize;		/*	number of rows	*/
	ushort		rowbytes;	/*	bytes per row in media (typically multiple of 4) */
/*	uchar		data[ysize][rowbytes];
 *	here, you write the data of the bitmap image
 */
};

/*	For the MEDIA_8_BIT_PALETTE MediaBitmapHeader (which is not really 
 *	recommended, but put in here due to popular pressure) you should precede 
 *	the first frame (and any other frame for which you want a change of palette) 
 *	with this header. The reader, when reading the stream and finding this chunk, 
 *	will immediately change the palette to the one specified there. Typically, it 
 *	should have the same "position" as the MediaBitmapHeader.
 */
enum { MEDIA_PALETTE_CHUNK = 'mpac' };
struct MediaPaletteHeader {
	ulong		magic;
	ulong		hdrsize;
	ulong		datasize;
	long		position;
	long		numcolors;
	ushort		space;		/*	tells order and format of colors - 32_BIT or 16_BIT values only	*/
/*	uchar		data[numcolors][entrysize(space)];
 *	here, you write the values for the palette. The palette will take effect for each MEDIA_BITMAP_CHUNK 
 *	frame after this.
 */
};

/*	MEDIA_SOUND_HEADER defines what the total sound track will look like.
*	It uses the same struct MediaSoundHeader as MEDIA_SOUND_CHUNK, but no 
*	data. This is an optional header to prepare the player for what'll come.
*/
enum { MEDIA_SOUND_HEADER = 'smnh' };
/*	MEDIA_SOUND_CHUNK contains actual sound data to play. Due to rounding 
 *	error in "position", any MEDIA_SOUND_CHUNK that appears to have a gap 
 *	of only one sample should be considered playable end-to-end with the previous 
 *	sound chunk.
 */
enum { MEDIA_SOUND_CHUNK = 'msnc' } ;
struct MediaSoundHeader {
	ulong		magic;		/*	MEDIA_SOUND_CHUNK	 */
	ulong		hdrsize;	/*	(24 for now)	*/
	ulong		datasize;	/*	divide by channels times padfunc(samplebits) to get number of frames	*/
	long		position;	/*	media clock when to play it	*/
	long		samplerate;	/*	44100 or 48000 mostly	*/
	short		channels;	/*	2 for stereo, left first	*/
							/*	same as "number samples per samplerate frame".	*/
	short		samplebits;	/*	bits per sample - stored in low order bits, padded 	*/
							/*	to even power of two sized samples	*/
							/*	8 -> 1 byte per sample	*/
							/*	12 -> 2 bytes per sample	*/
							/*	16 -> 2 bytes per sample	*/
							/*	24 -> 4 bytes per sample	*/
							/*	Always linear encoding	*/
/*	uchar		data[datasize/<framesize>][channels][padfunc(samplebits)]
 *	<framesize> = padfunc(samplebits)*channels
 *	here, you write sound data
 */
};

/*	For loops (and markers, really) you can emit this chunk. 
 *	A good place to do it is in immediate vincinity of the chunk that starts the loop
 *	If the loop starts within a chunk, the loop chunk should come after that chunk
 *	since otherwise "position" would not be monotonically increasing or flat.
 *	Readers should beware that a loop chunk may precede a data chunk it relates 
 *	to in the case when the loop starts where the data chunk starts...
 */
enum { MEDIA_LOOP_CHUNK = 'mlop' };
struct MediaLoopHeader {
	ulong		magic;			/*	MEDIA_LOOP_CHUNK	*/
	ulong		hdrsize;		/*	24 for now	*/
	ulong		datasize;		/*	strlen(loopname)+1	*/
	long		position;		/*	Where to start looping	*/
	long		end;			/*	Where to loop back from	*/
	long		loopcount;		/*	0 == user starts, no looping
								 *	1 == auto-start, no looping
								 *	2+ == auto-start, loop this many times
								 *	-1 == loop infinite
								 */
/*	char			loopname[datasize];
 *	zero-terminated loop name string
 */
};


enum { MEDIA_END_MAGIC = 'mbye' };
struct MediaStreamTrailer {
	ulong		magic;		/*	MEDIA_END_MAGIC	*/
	long		headersize;	/*	(20 for now)	*/
	long		datasize;	/*	size of text (not zero-terminated) that follows	*/
	long		endtime;	/*	media clock when file done	*/
	long		numframes;	/*	number of "major" frames; typically number of video 	*/
/*	char		textinfo[datasize];
 *	This is an arbitrarily-sized "comment" field where copyrights, human-readable information, 
 *	notes, ... whatever can be stored. It is not zero-terminated.
 */
};


#if defined(__cplusplus) && defined(__BEOS__)

#include <OS.h>			/*	for sem_id and thread_id	*/

class BBitmap;
class DStream;

/*	This simple utility class will write a standard baseline media stream for you.
 */
class MEDIAWriter {
public:
					MEDIAWriter(
						DStream &		outstream,
						long			samplerate = 44100,
						bool			usesvideo = true,
						bool			usessound = false);
					~MEDIAWriter();
	void			SetComment(
						const char *	comment);

	long			PutBitmap(
						BBitmap *		bitmap,
						long			afterSamples = 4410);	/*	delay BEFORE bitmap triggers	*/
	long			PutBitmapHeader(
						BRect			area,
						color_space		space,
						long			rowbytes);
	long			PutBitData(
						const void *	bits,
						int				rowBytes,
						int				xsize,
						int				ysize,
						int				numCols,
						int				numRows,
						color_space		space,
						long			afterSamples = 4410);	/*	delay BEFORE bitmap triggers	*/
	long			PutSound(
						const void *	data,
						long			numsamples,				/*	delay AFTER sound triggers	*/
						int				numchannels = 2,
						int				samplebits = 16);
	/*	This means that if you want to trigger a sound and a bitmap at the same time, you 
	 *	should write the bitmap first, then the sound.
	 */
protected:
	DStream &		fOutput;
	long			fSampleRate;
	long			fBitmapCount;
	long			fSoundCount;
	long			fLastBitmapClock;
	long			fNextSoundClock;
	long			fLastSoundClock;
	char *			fComment;
};


/*	MEDIANotifier is defined for use by MEDIAReaderStream.
 *	MEDIAReaderStream will call the various hooks in the MEDIANotifier passed to its 
 *	constructor as data arrives.
 */
class MEDIANotifier {
public:
					MEDIANotifier() { }
virtual				~MEDIANotifier() { }

virtual	void		BitmapCreated(
						BBitmap			*theBitmap) = 0;
virtual	void		BitmapUpdated(
						BBitmap			*theBitmap,
						int				xPos,
						int				yPos,
						int				xSize,
						int				ySize) = 0;
virtual	void		AudioSetup(
						long			samplerate,
						int				channels,
						int				bits) = 0;
virtual	void		AudioReceived(
						const void		*audioData,
						int				bufferSize,
						long			samplerate,
						int				channels,
						int				bits) = 0;
virtual	void		StreamEnded() = 0;
};

/*	DefaultNotifier is a stock object that is suitable to pass into MEDIAReaderStream. 
 *	It takes a view in which it will draw as argument, and it can optionally resize this 
 *	view once the size of the frame is known.
 *	It can also start the AudioSubscriber business all by itself; no user intervention 
 *	needed.
 */
class DefaultNotifier :
	public MEDIANotifier
{
		BView		*fView;
		bool		fResize;
		bool		fHasFinished;
public:
					DefaultNotifier(
						BView			*view,
						bool			resize = false,
						bool			wantAudio = false);
					~DefaultNotifier();

		bool		HasFinished();

virtual	void		BitmapCreated(
						BBitmap			*theBitmap);
virtual	void		BitmapUpdated(
						BBitmap			*theBitmap,
						int				xPos,
						int				yPos,
						int				xSize,
						int				ySize);
virtual	void		AudioSetup(
						long			samplerate,
						int				channels,
						int				bits);
virtual	void		AudioReceived(
						const void		*audioData,
						int				bufferSize,
						long			samplerate,
						int				channels,
						int				bits);
virtual	void		StreamEnded();
};

/*	This stream is useful when translating into DATA_VIDEO, if you want to immediately 
 *	"play" the data you're importing. Note that it doesn't handle translators that Seek() or 
 *	SetSize() the output stream, but those should be uncommon.
 */
class MEDIAReaderStream :
	public DStream
{
public:
						MEDIAReaderStream(
							MEDIANotifier	*notifier) throw(long);
						~MEDIAReaderStream();

		long			Read(	//	returns 0 for success
							void *			buffer,
							long &			size);
		long			Write(	//	returns 0 for success
							const void *	data,
							long			size);
		long			Seek(	//	returns 0 for success
							long			position);
		long			Position(); // returns position
		long			Size(); // returns total size
		long			SetSize(	//	returns 0 for success
							long			size);

		BBitmap			*DetachBitmap();

protected:

		void			*fData;
		int				fDataPhys;
		int				fDataLog;
		long			fPosition;
		double			fPlayBaseTime;
		double			fTicksPerSec;

		long			lastsamplerate;
		long			lastchannels;
		long			lastbits;
		long			bitmapx;
		long			bitmapy;
		BBitmap			*fBitmap;
		MEDIANotifier	*fNotifier;
		ulong			fPalette[256];

virtual	void			HandleChunk(					/*	override to handle chunks not handled by this	*/
							ulong			magic,		/*	chunk type out of header	*/
							const void		*header,	/*	header			*/
							const void		*data,		/*	data after header		*/
							int				dataSize);	/*	size of data block		*/
virtual	void			NapTime(						/*	snooze until header item is supposed to be played	*/
							const void		*header);	/*	you can override to, say, do free run, or alter playback speed	*/

private:

		/*	These are used by the implementation of the bitmap playback	*/
		void			ResizeBitmap(
							MediaBitmapHeader	*header);
		void			ParseBitmapData(
							MediaBitmapHeader	*header,
							const void			*data,
							int					datasize);
		void			InitPalette();	/*	Called from constructor	*/
		void			SetPalette(
							MediaPaletteHeader	*header,
							const void			*data,
							int					size);
};

#endif
