// TCPStream.cpp by Simon Clarke (S.J.Clarke@herts.ac.uk)

// This is a stream designed to read from a socket (writing is not
// supported yet) and cache any results so it can be prepared to
// seek around the buffer.  You can implement your own datastream
// class based on this very easily

// The seeking code is a bit of a nightmare, considering that it
// has to read up to the point you want to seek to.  However, it 
// does cache everything, so when it comes to reading the file
// again the delay will not be the same.

// Please feel free to use this within your project, as long as 
// you give me a small credit :)

#include "TCPStream.h"

TCPStream::TCPStream(char *sitename, unsigned short port)
{
	fConnected = FALSE;	
	if (fSocket = socket(AF_INET, SOCK_STREAM, 0) < 0) return;
	
	fCache = 0;
	fUseCache = TRUE;
	fPosition = 0;
	fCachePosition = 0;	 
	fSitename = sitename;
	fPort = port;
}

TCPStream::~TCPStream()
{
	if (fSocket != -1) closesocket(fSocket);
	if (fCache) free(fCache);
}

long TCPStream::Read(void *buffer, long &size)
{	
	if (fSocket == -1) return B_ERROR;
	if (!fConnected) {

		if (!(OurAdd = gethostbyname(fSitename))) return B_ERROR;
		
		bzero(&SockBlock, sizeof(SockBlock));
		bcopy(OurAdd->h_addr, (char *)&SockBlock.sin_addr, OurAdd->h_length);
		SockBlock.sin_family = OurAdd->h_addrtype;
		SockBlock.sin_port = htons(fPort); // TCP Port
				
		if (connect(fSocket,(struct sockaddr *) &SockBlock, sizeof(SockBlock)) < 0 ) {
			size = 0;
			return B_ERROR;
		} else {
			// send request to server
			fConnected = TRUE;
			
			if (SendRequest(fSocket) != B_NO_ERROR) {			
				size = 0;
				return B_ERROR;
			}
		}
	}
	char *fBuffer = (char *)buffer;
	long fSize = size;

	// can we and should we return the data from local cache

	if ((fCachePosition >= fPosition) && (fCachePosition) && 
	  (fCachePosition != fPosition) && (fUseCache)) {
	
		if ((fSize+fPosition > fCachePosition)) {
			
			long cacheread = (fPosition+fSize)-fCachePosition;
			if ((fPosition+fSize) > fCachePosition) cacheread = fCachePosition;

			memcpy(fBuffer,fCache+fPosition,cacheread);
			fBuffer += cacheread;
			fSize = fSize - cacheread;
			fPosition += cacheread;			
		} else {
			memcpy(fBuffer,fCache+fPosition,fSize);
			fPosition += fSize;
			return B_NO_ERROR;
		}
	}

	// keep polling the connection until all the required amount of data
	// has been read

	do {
		int dataread = recv(fSocket,fBuffer,fSize,NULL);
		fSize-=dataread;
	
		if (dataread > 0) {
			
			// place into local cache to allow backward seeking, allow cache
			// to be turned off, so streams can send and receive commands			
			if (fUseCache) {
				fCachePosition+=dataread;
										
				if (!fCache) {
					fCache = (char *)malloc(dataread);
					memcpy(fCache,fBuffer,dataread);
				} else {
	
					void *fNewCache = realloc(fCache,fCachePosition);
					if (!fNewCache) { // no memory
						size = 0;
						return B_NO_MEMORY;
					}
					fCache = (char *)fNewCache;
					memcpy((fCache+fCachePosition)-dataread,fBuffer,dataread);
				}
			}
			fPosition+=dataread;
			fBuffer+=dataread;
		
			// has everything needed been read?
			if (fSize == NULL) return B_NO_ERROR;
		}

		if (dataread == 0) {
			size = size - fSize;
			return B_NO_ERROR;
		}
		if (dataread == -1) {
			size = 0;
			return B_ERROR;
		}
	} while (1);	
}

long TCPStream::Size()
{
	return B_ERROR;
}

long TCPStream::Position()
{
	return fPosition;
}

long TCPStream::SetSize(long size)
{
	return B_ERROR;
}	

long TCPStream::Seek(long position)
{
	if (position < 0) return B_ERROR;
	if (position == 0) {
		fPosition = 0;
		return B_NO_ERROR;
	}

	if (position > fCachePosition) {
		long readbytes = position-fPosition;
		void *temp = malloc(readbytes);
		Read(NULL,readbytes); // dirty way of reading beyond current pos
		free(temp);
		return B_NO_ERROR;
	}	
		
	fPosition = position;
	return B_NO_ERROR;
}

long TCPStream::Write(const void *data, long size)
{
	return B_ERROR;
}

void TCPStream::CacheResume()
{
	fUseCache = TRUE;
}

void TCPStream::CachePause()
{
	fUseCache = FALSE;
}

bool TCPStream::IsCacheOn()
{
	return fUseCache;
}

long TCPStream::BytesRead()
{
	return fCachePosition;
}