//	DataStreams.cpp

#include "DataStreams.h"

#include <string.h>
#include <OS.h>
#include <stdlib.h>




#pragma mark --- DFileStream ---

DFileStream::DFileStream(
	BFile * file,
	long openMode,
	long offset)
{
	fFile = file;
	fOffset = offset;
	fOpenMode = openMode;
	if (file->IsOpen())
		file->Seek(offset, 0);
}


DFileStream::~DFileStream()
{
	if (fDispose) {
		if (fFile->IsOpen())
			fFile->Close();
		delete fFile;
	}
}


long
DFileStream::OpenFile()
{
	if (fFile->IsOpen())
		return B_NO_ERROR;
	long err = fFile->Open(fOpenMode);
	if (err >= 0)
		err = fFile->Seek(fOffset, 0);
	if (err == fOffset)
		err = 0;
	return err;
}


long
DFileStream::Read(
	void * buffer,
	long & size)
{
	long err = OpenFile();
	if (err < 0)
		return err;
	size = fFile->Read(buffer, size);
	if (size < 0)
		return size;
	if (!size)
		return B_ERROR;
	return 0;
}


long
DFileStream::Write(
	const void * buffer,
	long size)
{
	long err = OpenFile();
	if (err < 0)
		return err;
	err = fFile->Write(buffer, size);
	if (err < 0)
		return err;
	return 0;
}


long
DFileStream::Seek(
	long position)
{
	if (position < 0)
		return B_ERROR;
	long err = OpenFile();
	if (err < 0)
		return err;
	if (position > Size())
		return B_BAD_VALUE;
	err = fFile->Seek(position+fOffset, 0);
	if (err < 0)
		return err;
	return 0;
}


long
DFileStream::Position()
{
	long err = OpenFile();
	if (err < 0)
		return err;
	err = fFile->Seek(0, 1);
	if (err >= 0)
		err -= fOffset;
	return err;
}


long
DFileStream::Size()
{
	long err = OpenFile();
	if (err < 0)
		return err;
	long pos = fFile->Seek(0, 1);
	long size = fFile->Seek(0, 2);
	if (pos >= 0)
		fFile->Seek(pos, 0);
	return size-fOffset;
}


long
DFileStream::SetSize(
	long size)
{
	long err = OpenFile();
	if (err < 0)
		return err;
	long pos = fFile->Seek(0, B_SEEK_MIDDLE);
	fFile->Seek(size+fOffset, B_SEEK_TOP);
	err = fFile->SetSize(size+fOffset);
	if (pos > size+fOffset)
		pos = size+fOffset;
	if (pos >= 0)
		fFile->Seek(pos, B_SEEK_TOP);
	return err;
}





#pragma mark --- DMallocStream ---

DMallocStream::DMallocStream(
	void * ptr,
	long size,
	long offset)
{
	fPtr = ptr;
	if (!fPtr)
		fPtr = malloc(size+offset+1);	//	make sure we never get NULL
	fPhysSize = size+offset;
	fLogSize = size;
	fLogPos = 0;
	fOffset = offset;
}


DMallocStream::~DMallocStream()
{
	if (fDispose && fPtr) free(fPtr);
}


long
DMallocStream::Read(
	void * buffer, 
	long & size)
{
	if (fLogPos >= fLogSize) {
		size = 0;
		return B_ERROR;
	}
	if (size > fLogSize-fLogPos)
		size = fLogSize-fLogPos;
	memcpy(buffer, ((char *)fPtr)+fOffset+fLogPos, size);
	fLogPos += size;
	return 0;
}


long
DMallocStream::Write(
	const void * buffer,
	long size)
{
	if (size+fLogPos+fOffset > fPhysSize) {
		long newSize = fPhysSize + (fPhysSize-fOffset);
		if (newSize < size+fLogPos+fOffset)
			newSize = size+fLogPos+fOffset+1024;
		void *newPtr = realloc(fPtr, newSize);
		if (!newPtr)
			return B_NO_MEMORY;
		fPtr = newPtr;
		fPhysSize = newSize;
	}
	memcpy(((char *)fPtr)+fOffset+fLogPos, buffer, size);
	fLogPos += size;
	if (fLogPos > fLogSize)
		fLogSize = fLogPos;
	return 0;
}


long
DMallocStream::Seek(
	long position)
{
	if (position < 0 || position > fLogSize)
		return B_ERROR;
	fLogPos = position;
	return B_NO_ERROR;
}


long
DMallocStream::Position()
{
	return fLogPos;
}


long 
DMallocStream::Size()
{
	return fLogSize;
}


long
DMallocStream::SetSize(
	long size)
{
	if (fLogSize != size) {
		if (size > fPhysSize-fOffset) {
			void *newPtr = realloc(fPtr, size+fOffset);
			if (!newPtr)
				return B_NO_MEMORY;
			fPtr = newPtr;
			fPhysSize = size+fOffset;
		}
		fLogSize = size;
		if (fLogPos > size)
			fLogPos = size;
	}
	return B_NO_ERROR;
}


const void *
DMallocStream::Peek(
	long & outOffset,
	long & outSize)
{
	outOffset = fOffset;
	outSize = fLogSize;
	return fPtr;
}




#pragma mark --- DAreaStream ---

DAreaStream::DAreaStream(
	area_id area,
	long size,
	long offset)
{
	fPhysSize = size+offset;
	fLogSize = size;
	fLogPos = 0;
	fOffset = offset;
	fArea = area;
	if (!area) {
		long s = (size+offset+4095)&~4095;	//	page size is 4096
		area = create_area("DAreaStream", &fPtr, B_ANY_ADDRESS,
				s, B_NO_LOCK, B_READ_AREA | B_WRITE_AREA);
		if (area < 0)
			return;
		fArea = area;
	}
}


DAreaStream::~DAreaStream()
{
	if (fDispose && (fArea > 0))
		delete_area(fArea);
}


long
DAreaStream::Read(
	void * buffer, 
	long & size)
{
	if (fLogPos >= fLogSize) {
		size = 0;
		return B_ERROR;
	}
	if (size > fLogSize-fLogPos)
		size = fLogSize-fLogPos;
	memcpy(buffer, ((char *)fPtr)+fOffset+fLogPos, size);
	fLogPos += size;
	return 0;
}


long
DAreaStream::Write(
	const void * buffer,
	long size)
{
	if (size+fLogPos+fOffset > fPhysSize) {
		long newSize = fPhysSize + (fPhysSize-fOffset);
		if (newSize < size+fLogPos+fOffset)
			newSize = size+fLogPos+fOffset+1024;
		newSize = (newSize+4095)&~4095;
		long err = resize_area(fArea, newSize);
		if (err)
			return err;
		area_info info;
		get_area_info(fArea, &info);
		fPtr = info.address;
		fPhysSize = newSize;
	}
	memcpy(((char *)fPtr)+fOffset+fLogPos, buffer, size);
	fLogPos += size;
	if (fLogPos > fLogSize)
		fLogSize = fLogPos;
	return 0;
}


long
DAreaStream::Seek(
	long position)
{
	if (position < 0 || position > fLogSize)
		return B_ERROR;
	fLogPos = position;
	return B_NO_ERROR;
}


long
DAreaStream::Position()
{
	return fLogPos;
}


long 
DAreaStream::Size()
{
	return fLogSize;
}


long
DAreaStream::SetSize(
	long size)
{
	if (fLogSize != size) {
		if (size > fPhysSize+fOffset) {
			long newSize = (size+fOffset+4095)&~4095;
			long err = resize_area(fArea, newSize);
			if (err)
				return err;
			area_info info;
			get_area_info(fArea, &info);
			fPtr = info.address;
			fPhysSize = newSize;
		}
		fLogSize = size;
		if (fLogPos > size)
			fLogPos = size;
	}
	return B_NO_ERROR;
}


const void *
DAreaStream::Peek(
	long & outOffset,
	long & outSize)
{
	outOffset = fOffset;
	outSize = fLogSize;
	return fPtr;
}

