//	Datatypes.cpp

#include "DStream.h"
#include "Datatypes.h"
#include "DataHandler.h"
#include "DList.h"
#if PROTECTION
	#include "Acquire.h"
#endif

#include <OS.h>
#include <image.h>
#include <File.h>
#include <Directory.h>
#include <Query.h>
#include <Volume.h>
#include <Database.h>
#include <string.h>
#include <OS.h>
#include <stdlib.h>



const char dataTranslatorFolders[]		= "TranslatorFolders";
const char dataSettingsTableName[]		= "DatatypesSettings";
const char dataSettingsKeyName[]		= "Key";
const char dataSettingsDataName[]		= "Data";
const char dataDefaultTranslatorPath[]	= "/boot/Datatypes";


struct XLInfo {
					XLInfo() { memset(this, 0, sizeof(*this)); }

	XLInfo *		next;

	record_ref		ref;
	image_id		image;
	const char *	handlerName;
	const char *	handlerInfo;
	long			handlerVersion;
	const Format *	inputFormats;
	const Format *	outputFormats;

	long			(*Identify)(
						DStream &			inSource,
						const Format *		inFormat,
						BMessage *			ioExtension,
						DATAInfo &			outInfo,
						ulong				outType);
	long			(*Translate)(
						DStream &			inSource,
						const DATAInfo &	inInfo,
						BMessage *			ioExtension,
						ulong				outFormat,
						DStream &			outDestination);
	long			(*MakeConfig)(
						BMessage *			ioExtension,
						BView * &			outView,
						BRect &				outExtent);
};


class StAcquire {
	sem_id		fSem;
public:
				StAcquire(
					sem_id sem)
				{
					acquire_sem(sem);
					fSem = sem;
				}
				~StAcquire()
				{
					release_sem(fSem);
				}
};


static XLInfo *sInfo;
static sem_id sSem;


const char *
DATAVersion(	//	returns version string
	long & outCurVersion,	//	current version spoken
	long & outMinVersion)	//	minimum version understood
{
	static char vString[50];
	static char vDate[] = __DATE__;
	if (!vString[0]) {
		sprintf(vString, "Datatypes Library v%d.%d.%d %s\n",
			DATA_CURRENT_VERSION/100,
			(DATA_CURRENT_VERSION/10)%10,
			DATA_CURRENT_VERSION%10,
			vDate);
	}
	outCurVersion = DATA_CURRENT_VERSION;
	outMinVersion = DATA_MIN_VERSION;
	return vString;
}


static long
LoadTranslator(
	record_ref ref)
{
	//	check that this ref is not already loaded
	bool found = FALSE;
	for (XLInfo *i = sInfo; i && !found; i=i->next)
		if (i->ref == ref)
			found = TRUE;
	if (found)
		return B_NO_ERROR;

	//	sanity-check the file
	BFile file(ref);
	long err;
	if (0 != (err = file.Error()))
		return err;

	ulong type, app;
	if (0 != (err = file.GetTypeAndApp(&type, &app)))
		return err;

	if ((type != 'BAPP') && (type != DATAHANDLER_TYPE))
		return B_NO_ERROR;

	//	go ahead and load it
	image_id image = load_add_on(&file);
	if (image < 0)
		return image;

	XLInfo *info = new XLInfo;
	info->ref = ref;
	info->image = image;

	//	find all the symbols
	err = get_image_symbol(image, "handlerName", B_SYMBOL_TYPE_DATA,
			(void **)&info->handlerName);
	if (!err && get_image_symbol(image, "handlerInfo",
			B_SYMBOL_TYPE_DATA, (void **)&info->handlerInfo))
		info->handlerInfo = NULL;
	if (!err)
		err = get_image_symbol(image, "handlerVersion",
				B_SYMBOL_TYPE_DATA, (void **)&info->handlerVersion);
	if (!err && get_image_symbol(image, "inputFormats",
			B_SYMBOL_TYPE_DATA, (void **)&info->inputFormats))
		info->inputFormats = NULL;
	if (!err && get_image_symbol(image, "outputFormats",
			B_SYMBOL_TYPE_DATA, (void **)&info->outputFormats))
		info->outputFormats = NULL;
	if (!err)
		err = get_image_symbol(image, "Identify", B_SYMBOL_TYPE_TEXT,
				(void **)&info->Identify);
	if (!err)
		err = get_image_symbol(image, "Translate",
				B_SYMBOL_TYPE_TEXT, (void **)&info->Translate);
	if (!err && get_image_symbol(image, "MakeConfig",
			B_SYMBOL_TYPE_TEXT, (void **)&info->MakeConfig))
		info->MakeConfig = NULL;

	if (err) {	//	this is not a correct add-on
		delete info;
		return err;
	}

#if DEBUG
	{
		char name[B_FILE_NAME_LENGTH];
		file.GetName(name);
	}
#endif

	//	add to global list
	info->next = sInfo;
	sInfo = info;

	return B_NO_ERROR;
}


static void
LoadDir(
	record_ref ref,
	long & loadErr,
	long & nLoaded)
{
	BDirectory dir(ref);
	long err;
	if (B_NO_ERROR != (err = dir.Error())) {
		loadErr = err;
		return;
	}

	int fileCount = dir.CountFiles();
	if (fileCount > 0) {
		record_ref *refs = new record_ref[fileCount];
		err = dir.GetFiles(0, fileCount, refs);
		if (err)
			loadErr = err;
		else for (int ix=0; ix<fileCount; ix++) {
			err = LoadTranslator(refs[ix]);
			if (err)
				loadErr = err;
			else
				nLoaded++;
		}
		delete[] refs;
	}
}


long
DATAInit(		//	establish connection
	ulong /* mySignature */,
	BQuery * translators)
{
	if (sSem <= 0)
		sSem = create_sem(1, "Datatypes Lock");
	if (sSem <= 0)
		return sSem;

	StAcquire lock(sSem);

	long loadErr = 0;
	long nLoaded = 0;

	if (translators) {

		//	build a list of records that we want to load
		database_id db;
		DList<record_id> records;

		db = translators->Database()->ID();
		if (db < 0)
			return db;	//	error !

		long max = translators->CountRecordIDs();
		for (int ix=0; ix<max; ix++) {

			record_id id = translators->RecordIDAt(ix);
			if (id < 0)
				loadErr = id;	//	error!
			else
				records.AddItem(id);
		}

		//	load from one database
		for (int ix=0; ix<records.CountItems(); ix++) {
	
			//	extract the item
			record_ref ref;
			ref.database = db;
			ref.record = records.ItemAt(ix);
	
			long err = LoadTranslator(ref);
			if (err)
				loadErr = err;
			else
				nLoaded++;
		}
	} else {

		//	load whatever the defaults are

		//	get preferences from database
		long size;
		record_ref *dref;
		long err = DATAGetSetting(dataTranslatorFolders, (void *&)dref, size);

		//	if that fails, try the default path
		if (!err && dref) {
			size = size / sizeof(record_ref);

			//	if we have nothing to load, just cop out
			if (err || !size) {
				free(dref);
				return err;
			}

			//	for each folder record_ref in the list, load its files
			for (int iy=0; iy<size; iy++) {
				LoadDir(dref[iy], loadErr, nLoaded);
			}
			free(dref);
		}
		//	now load environment variable
		const char *dtpath = getenv("DATATYPES");
		if (!dtpath)
			dtpath = dataDefaultTranslatorPath;
		record_ref tref;
		err = get_ref_for_path(dtpath, &tref);
		if (!err) {
			LoadDir(tref, loadErr, nLoaded);
		}
	}

	if (nLoaded)
		loadErr = B_NO_ERROR;
	return loadErr;
}


long
DATAShutdown()	//	don't want to talk anymore
{
	if (sSem > 0) {

		acquire_sem(sSem);

		XLInfo *info = sInfo;
		sInfo = NULL;
		while (info != NULL) {
			XLInfo *del = info;
			info = info->next;
			unload_add_on(del->image);
			delete del;
		}

		delete_sem(sSem);
		sSem = 0;

	} else {

		return DATA_NOT_INITIALIZED;	
	}
	return B_NO_ERROR;
}


//	these functions call through to the translators


//	CheckFormats is a utility function that returns TRUE if the 
//	data provided and translator info can be used to make a 
//	determination (even if that termination is negative) and 
//	FALSE if content identification has to be done.
static bool
CheckFormats(
	XLInfo *info,
	ulong hintType,
	const char * hintMIME,
	const Format * & outFormat)
{
	outFormat = NULL;

	//	return FALSE if we can't use hints for this module
	//
	if (!hintType && !hintMIME)
		return FALSE;
	if (!info->inputFormats)
		return FALSE;

	//	scan for suitable format
	//
	const Format *fmt = info->inputFormats;
	int mlen = 0;
	//	check for the length of the MIME string, since it may be just a prefix
	//	so we use strncmp().
	if (hintMIME)
		mlen = strlen(hintMIME);
	while (fmt->type) {
		if ((fmt->type == hintType) ||
				(hintMIME && mlen && !strncmp(fmt->MIME, hintMIME, mlen)))
		{
			outFormat = fmt;
			return TRUE;
		}
		fmt++;
	}
	//	the module did export formats, but none mathced.
	//	we return true (uses formats) but set outFormat to NULL
	return TRUE;
}


long
DATAIdentify(	//	find out what something is
	DStream &			inSource,
	BMessage *			ioExtension,
	DATAInfo &			outInfo,
	ulong				inHintType,
	const char *		inHintMIME,
	ulong				inWantType)
{
	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	XLInfo *info = sInfo;
	XLInfo *bestMatch = NULL;
	float bestWeight = 0.0;
	while (info) {

		Format *format = NULL;
		DATAInfo tmpInfo;
		float weight = 0.0;

		long err = inSource.Seek(0);
		if (err) {
			return err;
		}

		if (CheckFormats(info, inHintType, inHintMIME, format)) {

			//	after checking the formats for hints, we still need to make sure the translator recognizes 
			//	the data and can output the desired format, so we call its' Identify() function.
			if (format && !info->Identify(inSource, format, ioExtension,
					tmpInfo, inWantType)) {

				goto add_match;
			}
		} else if (!info->Identify(inSource, NULL, ioExtension, 
				tmpInfo, inWantType)) {

	add_match:
			 // XOR to discourage taking advantage of undocumented features.
			tmpInfo.formatHandler = info->image^DATA_CURRENT_VERSION;
			weight = tmpInfo.formatQuality * tmpInfo.formatCapability;

			if (weight > bestWeight) {

				bestMatch = info;
				bestWeight = weight;
				outInfo = tmpInfo;
			}
		}
		info = info->next;
	}
	return bestMatch ? B_NO_ERROR : DATA_NO_HANDLER;
}


	static int
	compare_data(
		const void *a,
		const void *b)
	{
		register DATAInfo *ai = *(DATAInfo **)a;
		register DATAInfo *bi = *(DATAInfo **)b;
		return - ai->formatQuality*ai->formatCapability +
				bi->formatQuality*bi->formatCapability;
	}

long
DATAGetHandlers(	//	find all handlers for a type
	DStream &			inSource,
	BMessage *			ioExtension,
	DATAInfo * &		outInfo,	//	call delete[] on outInfo when done
	long &				outNumInfo,
	ulong				inHintType,
	const char *		inHintMIME,
	ulong				inWantType)
{
	outInfo = NULL;
	outNumInfo = 0;

	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	long physCnt = 10;
	outInfo = new DATAInfo[physCnt];
	outNumInfo = 0;

	XLInfo *info = sInfo;

	while (info) {

		Format *format = NULL;
		DATAInfo tmpInfo;

		long err = inSource.Seek(0);
		if (err) {
			delete outInfo;
			outInfo = NULL;
			return err;
		}

		if (CheckFormats(info, inHintType, inHintMIME, format)) {
			if (format && !info->Identify(inSource, format, ioExtension, tmpInfo, 
					inWantType)) {

				goto add_match;
			}
		} else if (!info->Identify(inSource, NULL, ioExtension, tmpInfo, inWantType)) {

	add_match:
			//	dynamically resize output list
			//
			if (physCnt <= outNumInfo) {

				physCnt += 10;
				DATAInfo *nOut = new DATAInfo[physCnt];
				for (int ix=0; ix<outNumInfo; ix++) {

					nOut[ix] = outInfo[ix];
				}
				delete[] outInfo;
				outInfo = nOut;
			}

			 // XOR to discourage taking advantage of undocumented features
			tmpInfo.formatHandler = info->image^DATA_CURRENT_VERSION;

			outInfo[outNumInfo++] = tmpInfo;
		}
		info = info->next;
	}
	if (outNumInfo > 1) {
		qsort(outInfo, outNumInfo, sizeof(*outInfo), compare_data);
	}
	return outNumInfo ? B_NO_ERROR : DATA_NO_HANDLER;
}


extern long
DATAGetAllHandlers(
	DATAID * & outList,
	long & outCount)
{
	outList = NULL;
	outCount = 0;

	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	//	count handlers
	XLInfo *info = NULL;
	for (info = sInfo; info != NULL; info = info->next)
		outCount++;
	outList = new DATAID[outCount];
	outCount = 0;
	for (info = sInfo; info != NULL; info = info->next)
		outList[outCount++] = info->image^DATA_CURRENT_VERSION;

	return B_NO_ERROR;
}


extern long
DATAGetHandlerInfo(
	DATAID				forHandler,
	const char * &		outName,
	const char * &		outInfo,
	long &				outVersion)
{
	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	//	find the handler we've requested
	XLInfo *info = NULL;
	for (info = sInfo; info != NULL; info = info->next) {

		if ((info->image^DATA_CURRENT_VERSION) == forHandler)
			break;
	}
	if (!info)
		return DATA_NO_HANDLER;

	outName = info->handlerName;
	outInfo = info->handlerInfo;
	outVersion = info->handlerVersion;

	return B_NO_ERROR;
}


extern long
DATAGetInputFormats(
	DATAID				forHandler,
	const Format * &	outFormats,		//	don't write contents!
	long &				outNumFormats)
{
	outFormats = NULL;
	outNumFormats = 0;

	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	//	find the handler we've requested
	XLInfo *info = NULL;
	for (info = sInfo; info != NULL; info = info->next) {

		if ((info->image^DATA_CURRENT_VERSION) == forHandler)
			break;
	}
	if (!info)
		return DATA_NO_HANDLER;

	outFormats = info->inputFormats;
	if (outFormats)
		while (outFormats[outNumFormats].type)
			outNumFormats++;

	return B_NO_ERROR;
}


extern long
DATAGetOutputFormats(
	DATAID forHandler,
	const Format * & outFormats,
	long & outNumFormats)
{
	outFormats = NULL;
	outNumFormats = 0;

	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	//	find the handler we've requested
	XLInfo *info = NULL;
	for (info = sInfo; info != NULL; info = info->next) {

		if ((info->image^DATA_CURRENT_VERSION) == forHandler)
			break;
	}
	if (!info)
		return DATA_NO_HANDLER;

	outFormats = info->outputFormats;
	if (outFormats)
		while (outFormats[outNumFormats].type)
			outNumFormats++;

	return B_NO_ERROR;
}


long
DATATranslate(	//	morph data into form we want
	DStream &			inSpec,
	const DATAInfo *	inInfo,			//	may be NULL
	BMessage *			ioExtension,
	DStream &			outSpec,
	ulong 				inWantOutType,
	ulong				inHintType,
	const char *		inHintMIME)
{
	DATAInfo stat_info;

	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	if (!inInfo) {	//	go look for a suitable translator?

		inInfo = &stat_info;

		long err = DATAIdentify(inSpec, ioExtension, stat_info, 
				inHintType, inHintMIME, inWantOutType);
		if (err)
			return err;
	}

	XLInfo *info = NULL;
	for (info = sInfo; info != NULL; info = info->next) {

		if ((info->image^DATA_CURRENT_VERSION) == inInfo->formatHandler)
			break;
	}
	if (!info)
		return DATA_NO_HANDLER;

	long err = inSpec.Seek(0);
	if (err) {
		return err;
	}

	return info->Translate(inSpec, *inInfo, ioExtension, inWantOutType, outSpec);
}


//	A translator may support settings for things such as preferred bit depth, etc.
//	Create a BView that contains controls and logic to "run" and change these settings.

long
DATAMakeConfig(
	DATAID				forHandler,
	BMessage *			ioExtension,
	BView * &			outView,
	BRect &				outExtent)
{
	if (sSem <= 0)
		return DATA_NOT_INITIALIZED;

#if PROTECTION
	Acquire lock(sSem);
#endif

	XLInfo *info = NULL;
	for (info = sInfo; info != NULL; info = info->next) {

		if ((info->image^DATA_CURRENT_VERSION) == forHandler)
			break;
	}
	if (!info || !info->MakeConfig)
		return DATA_NO_HANDLER;

	return info->MakeConfig(ioExtension, outView, outExtent);
}


static BTable *
CreateTable(
	BDatabase &			db)
{
	BTable *ret = db.CreateTable(dataSettingsTableName);
	if (!ret)
		return NULL;
	ret->AddStringField(dataSettingsKeyName, 0);
	ret->AddRawField(dataSettingsDataName, 0);

	//	Fill in default data
	record_ref ref;
	const char *dtpath = getenv("DATATYPES");
	if (!dtpath)
		dtpath = dataDefaultTranslatorPath;
	long err = get_ref_for_path(dtpath, &ref);
	if (!err && does_ref_conform(ref, "Folder")) {
		BRecord *search = new BRecord(ret);
		search->Lock();
		search->SetString(dataSettingsKeyName, dataTranslatorFolders);
		search->SetRaw(dataSettingsDataName, &ref, sizeof(ref));
		search->Commit();
		search->Unlock();
		delete search;
	}
	return ret;
}


long
DATAGetSetting(
	const char *		inSettingName,
	void * &			outSetting,
	long &				outSize)
{
	outSetting = 0;
	outSize = 0;
	BVolume vol = boot_volume();
	BTable *table = vol.Database()->FindTable(dataSettingsTableName);
	if (!table)
		table = CreateTable(*vol.Database());
	if (!table)
		return B_STORAGE_ERROR_BASE;	//	why is there no better error for this?
	BQuery *q = new BQuery;
	q->AddTable(table);
	q->PushString(inSettingName);
	q->PushField(dataSettingsKeyName);
	q->PushOp(B_EQ);
	q->FetchOne();
	if (!q->CountRecordIDs()) {
		delete q;
		return B_NAME_NOT_FOUND;
	}
	BRecord rec(vol.Database(), q->RecordIDAt(0));
	delete q;
	void *d = rec.FindRaw(dataSettingsDataName, &outSize);
	outSetting = malloc(outSize);
	memcpy(outSetting, d, outSize);
	return B_NO_ERROR;
}


long
DATASetSetting(
	const char *		inSettingName,
	const void *		inSetting,
	long				inSize)
{
	BVolume vol = boot_volume();
	BTable *table = vol.Database()->FindTable(dataSettingsTableName);
	if (!table)
		table = CreateTable(*vol.Database());
	if (!table)
		return B_STORAGE_ERROR_BASE;	//	why is there no better error for this?
	BQuery *q = new BQuery;
	q->AddTable(table);
	q->PushString(inSettingName);
	q->PushField(dataSettingsKeyName);
	q->PushOp(B_EQ);
	q->FetchOne();
	BRecord *theRec = NULL;
	if (!q->CountRecordIDs()) {
		theRec = new BRecord(table);
		theRec->Lock();
		theRec->SetString(dataSettingsKeyName, inSettingName);
	} else {
		theRec = new BRecord(vol.Database(), q->RecordIDAt(0));
		theRec->Lock();
	}
	delete q;
	theRec->SetRaw(dataSettingsDataName, inSetting, inSize);
	long err = theRec->Commit();
	theRec->Unlock();
	delete theRec;

	return err;
}





